#
# Smash product resolutions
#
# Copyright (C) 2009 Christian Nassau <nassau@nullhomotopie.de>
#
#  $Id: smasher.tcl,v 1.12 2009/05/31 10:59:53 cn Exp $
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#


oo::class create yacop::sagemodule {
    variable rescfg db actRMap diffMap
    constructor {rcfg dbcmd} {
        array set rescfg $rcfg
        set db $dbcmd

        my db eval {
            
            create temporary table if not exists 
            sagemod_gens(rowid integer primary key,
                         sagename text unique,
                         sdeg integer,
                         ideg integer,
                         edeg integer);
            
            create temporary table if not exists 
            sagemod_ops(type text,
                        op text,
                        srcgen integer,
                        result text);
            
        }

    }
    method db {subcmd args} {
        uplevel 1 [list $db $subcmd] $args
    }
  
    method addModuleGens {modgens} {
        # insert new generators into table
        set mcnt -1
        foreach ginfo $modgens {
            foreach {name degs} $ginfo break
            set gnum($name) [incr mcnt]
            set gdegs($name) $degs
        }
        my db eval {select sagename from sagemod_gens} {
            unset -nocomplain gdegs($sagename)
        }
        foreach {g degs} [array get gdegs] {
            dict with degs {
                my db eval {
                    insert into sagemod_gens(sagename,sdeg,ideg,edeg)
                    values($g,$s,$i,$e)
                }
            }
        }
        # identify missing information
        set todo(diff) {}
        my db eval {
            select sagename from sagemod_gens 
            where rowid not in (select srcgen from sagemod_ops where type = 'diff')
        } {
            lappend todo(diff) $gnum($sagename)
        }

        A configure -genlist {0} ;# a single generator in dim 0 
        unset -nocomplain aux
        if {$rescfg(viewtype) eq "even"} {
            set opsym Sq
        } else {
            set opsym Q
        }
        set p $rescfg(prime)
        set adict {}
        foreach newgen [array names gdegs] {
            set oplist {}
            set num $gnum($newgen) 
            my db eval {
                select distinct gt.ideg-gs.ideg as i, gt.edeg-gs.edeg as e
                from sagemod_gens as gs, sagemod_gens as gt
                where gs.sagename = $newgen and gt.sdeg = gs.sdeg 
                and i >= 0 and e >= 0
            } {
                A configure -ideg $i -edeg $e
                foreach m [A basis] {
                    lappend oplist (\"$m\",[yacop::sage::sagepoly $opsym false $p [list $m]])
                }
            }
            lappend adict $num:\[[join $oplist ,]\]
        }
        set todo(actR) [array get aux]
        return [format {{"diff":[%s],"actR":{%s}}} [join $todo(diff) ,] [join $adict ,]]
    }

    method addDiff {stuff} {
        #puts df=$stuff

        array set name2id [my db eval {
            select sagename,rowid from sagemod_gens
        }]

        foreach {gen diff} $stuff {
            set targ {}
            foreach {cf tg} $diff {
                lappend targ $cf $name2id($tg)
            }
            set sgen $name2id($gen)
            my db eval {
                insert into sagemod_ops(type,op,srcgen,result)
                values('diff',NULL,$sgen,$targ)
            }
        }
    }
    method addActionR {stuff} {
        #puts ar=$stuff
        array set name2id [my db eval {
            select sagename,rowid from sagemod_gens
        }]
        foreach {gen opspec} $stuff {
            set sgen $name2id($gen)
            foreach {op diff} $opspec {
                set targ {}
                foreach {cf tg} $diff {
                    lappend targ $cf $name2id($tg)
                }
                my db eval {
                    insert into sagemod_ops(type,op,srcgen,result)
                    values('actR',$op,$sgen,$targ)
                }
            }
        }
    }
    method generators {region} {
        set res {}
        my db eval {
            select sagename as name, sdeg, ideg, edeg from sagemod_gens
        } vals {
            unset -nocomplain vals(*)
            lappend res [array get vals]
        }
        return $res
    }
    method init {} {
        steenrod::monomap [set actRMap [namespace current]::actR]
        my db eval {
            select op,srcgen,result from sagemod_ops
            where type = 'actR'
        } {
            lset op end $srcgen
            $actRMap set $op $result
        } 
        #puts actRMap=[$actRMap list]
        unset -nocomplain diffMap
        my db eval {
            select srcgen,result from sagemod_ops
            where type = 'diff'
        } {
            set diffMap($srcgen) $result
        } 
    } 
    method actR {gen op} {
        array set res {}
        set p $rescfg(prime)
        steenrod::poly foreach $op mono {
            lset mono end $gen
            foreach {cf tar} [$actRMap get $mono] {
                incr res($tar) [expr {$cf*[steenrod::mono coeff $mono]}]
            } 
        }
        set rval {}
        foreach {key val} [array get res] {
            if {[set val [expr {$val % $p}]]} {
                lappend rval $val $key
            } 
        }
        return $rval
    }
    method diff {gen} {
        return $diffMap($gen) 
    }
}



oo::class create yacop::smashres {

    variable db res rescfg module

    constructor {resolution modulename modcmd {filename :memory:}} {
        set res $resolution
        set modname $modulename
        set module $modcmd

        yacop::sqlite [set db [namespace current]::db] $filename
 
        my db eval {
            create table if not exists yacop_cfg
            (key text unique, value text);

            insert or ignore into yacop_cfg(key,value)
            values('type','module_smash_gfr');
            insert or ignore into yacop_cfg(key,value)
            values('module',$modname);
            
            create table if not exists 
            module_generators(rowid integer primary key,
                              sagename text unique,
                              sdeg integer,
                              ideg integer,
                              edeg integer);

            create table if not exists 
            smash_generators(rowid integer primary key,
                             resgen integer,
                             modgen integer,
                             sdeg integer,
                             ideg integer,
                             edeg integer,
                             ebasid integer /* unique for given s,i,e */,
                             basid integer  /* unique for given s,i */,
                             havediff integer /* have we computed diffs on this generator? */);

            create unique index if not exists smash_generatorsIdx1
            on smash_generators(resgen,modgen);

            create unique index if not exists smash_generatorsIdx2
            on smash_generators(sdeg,ideg,edeg,ebasid);

            create unique index if not exists smash_generatorsIdx3
            on smash_generators(sdeg,ideg,basid);

            create trigger smash_generators_mkbasid
            after insert on smash_generators when new.basid is null
            begin
            update smash_generators set basid = (select count(*)-1 from smash_generators
                                                     where sdeg=new.sdeg and ideg=new.ideg)
            where rowid=new.rowid;
            end;

            create table if not exists 
            smash_fragments(rowid integer primary key,
                            /* this table contains only differentials 
                            .  and Pts actions - it is not a complete 
                            .  description of the smash product differential */
                            srcgen integer,
                            targen integer,
                            opideg integer,
                            format text,
                            data text);


            create table if not exists 
            smash_boxes(rowid integer primary key,
                        sdeg integer,
                        ideg integer,
                        edeg integer,
                        matrix text,
                        cycles text,
                        boundaries text,
                        homology text);

            create unique index if not exists 
            smash_boxes_idx1 on smash_boxes(sdeg,ideg,edeg);

            create table if not exists 
            homology_gendata(rowid integer primary key,
                             boxid integer,
                             ebasid integer /* unique for given s,i,e */,
                             basid integer  /* unique for given s,i */);
           
            create trigger homology_gendata_mkbasid
            after insert on homology_gendata when new.basid is null
            begin
            update homology_gendata set basid = (select count(*)-1 from homology_gendata
                                                 join smash_boxes as b on boxid=b.rowid
                                                 join smash_boxes as bn on new.boxid=bn.rowid
                                                 where b.sdeg=bn.sdeg and b.ideg=bn.ideg)
            where rowid=new.rowid;
            end;
            
            create unique index if not exists homology_gendata_idx1 
            on homology_gendata(boxid,ebasid);

            create view if not exists homology_generators as
            select h.rowid as rowid, b.sdeg as sdeg, b.ideg as ideg, b.edeg as edeg, 
            h.basid as basid, h.ebasid as ebasid
            from homology_gendata as h join smash_boxes as b on h.boxid=b.rowid;

            create table if not exists 
            homology_fragments(srcgen integer,
                               targen integer,
                               opideg integer,
                               format text,
                               data text);

            create index if not exists 
            homology_fragments_idx1 on homology_fragments(srcgen,opideg);

            create table if not exists chart_viewtypes
            (name text unique, desc text, code text);

        }

        foreach {key value} [$res db eval {
            select key,value from yacop_cfg
            where key in ('prime','algebra')
        }] {
            my db eval {
                insert or ignore into yacop_cfg(key,value) values($key,$value)
            }
        }
        array set rescfg [$res db eval {
            select key,value from yacop_cfg
        }]
        array set mycfg [my db eval {
            select key,value from yacop_cfg
        }]

        if {$mycfg(prime) != $rescfg(prime)
            || $mycfg(algebra) ne $rescfg(algebra)} {
            error "prime/algebra mismatch"
        }
        if {$mycfg(module) != $modulename} {
            error "file has modulename '$mycfg(module)' but module is '$modulename'"
        }
                       
        if {$rescfg(viewtype) eq "even"} {
            proc i2n {sdeg ideg} {expr {($ideg/2)-$sdeg}}
            proc n2i {sdeg ndeg} {expr {2*($ndeg+$sdeg)}}
        } else {
            proc i2n {sdeg ideg} {expr {$ideg-$sdeg}}
            proc n2i {sdeg ndeg} {expr {$ndeg+$sdeg}}
        }
        my db function i2n [namespace current]::i2n
        my db function n2i [namespace current]::n2i

        foreach {major minor} {
            smash-product even
            smash-product odd
            homology even
            homology odd
        } {
            if {[regexp even $minor] && ($rescfg(prime) != 2 || [lindex $rescfg(algebra) 0] != 0)} continue
            if {$rescfg(prime) != 2} {
                set name $major
            } else {
                set name "$major ($minor)"
            }
            set desc $name
            array set majorcode {
                smash-product {
                    create temporary view chart_generators as
                    select rowid, sdeg, IDEG-sdeg as ndeg, ideg, edeg, basid from smash_generators where havediff != 0;
                    create temporary view chart_fragments as select * from smash_fragments as sf join smash_generators as sg on sf.targen=sg.rowid where sg.havediff != 0;
                }
                homology {
                    drop view if exists chart_generators_aux;
                    create temporary view chart_generators_aux as 
                    select h.rowid,b.sdeg,b.ideg,b.edeg,h.basid from homology_gendata as h
                    join smash_boxes as b on h.boxid=b.rowid; 
                    create temporary view chart_generators as
                    select rowid, sdeg, IDEG-sdeg as ndeg, ideg, edeg, basid from chart_generators_aux;
                    create temporary view chart_fragments as select * from homology_fragments;
                }
            }
            if {[regexp even $minor]} {
                set map {IDEG ideg/2}
            } else {
                set map {IDEG ideg}
            }
            set code [string map $map $majorcode($major)]
            my db eval {
                insert or replace into chart_viewtypes(name,desc,code) values($name,$desc,$code);
            }
        }

        my db eval {
            insert or ignore into yacop_cfg(key,value) 
            values('viewtype',(select name from chart_viewtypes 
                               where name like '%homology%' order by name limit 1))
        }
        my db eval [my db onecolumn {select code from chart_viewtypes 
            where name=(select value from yacop_cfg where key='viewtype')}]

        steenrod::enumerator A -prime $mycfg(prime) -algebra [steenrod::algebra tomono $mycfg(algebra)]
    }

    method db {args} {
        if {[llength $args] == 0} {
            return $db
        }
        uplevel 1 [list $db] $args
    }
    method module {subcmd args} {
        uplevel 1 [list $module $subcmd] $args
    }

    method expand-region {region} {
        array set r $region
        foreach {nm off} {
            smax +1 smin -1
        } {
            if {[info exists r($nm)]} {
                incr r($nm) $off
            }
        }
        array get r
    }

    method compute {region} {
        $module init

        array set reg [my expand-region $region]
        my db transaction {
            foreach geninfo [my module generators $region] {
                dict with geninfo {
                    my db eval {
                        insert or ignore into
                        module_generators(sagename,sdeg,ideg,edeg)
                        values($name,$sdeg,$ideg,$edeg)
                    }
                }
            }
        }
        
        my db transaction {

            set conds {}
            foreach {bnd val} [array get reg] {
                switch -- $bnd {
                    smax {lappend conds " (SDEG)<=$val"}
                    imax {lappend conds " (IDEG)<=$val"}
                    smin {lappend conds " (SDEG)>=$val"}
                    imin {lappend conds " (IDEG)>=$val"}
                    nmax {
                        if {$rescfg(viewtype) eq "even"} {
                            lappend conds " (IDEG)<=(SDEG)+2*$val"
                        } else {
                            lappend conds " (IDEG)<=(SDEG)+$val"
                        }
                    }
                    nmin {
                        if {$rescfg(viewtype) eq "even"} {
                            lappend conds " (IDEG)>=(SDEG)+2*$val"
                        } else {
                            lappend conds " (IDEG)>=(SDEG)+$val"
                        }
                    }
                }
            } 
            if {$conds eq ""} {
                error "no restrictions given"
            }

            set aconds [join $conds { and }]
            set wherecond "where $aconds"
            set wherecond2 "where ([string map {SDEG SDEG0 IDEG IDEG0} $aconds]) or ([string map {SDEG SDEG1 IDEG IDEG1} $aconds]) or ([string map {SDEG SDEG2 IDEG IDEG2} $aconds])"
            set sql [string map {
                SDEG0 {$rsdeg+msdeg}   IDEG0 {$rideg+mideg}
                SDEG1 {$rsdeg+msdeg+1} IDEG1 {$rideg+mideg}
                SDEG2 {$rsdeg+msdeg-1} IDEG2 {$rideg+mideg}
            } [subst {
                select rowid as modgen, sdeg as msdeg, ideg as mideg, edeg as medeg
                from module_generators $wherecond2
            }]]
            #puts sql=$sql

            set regcond {select case when}
            append regcond [string map {SDEG $sdeg-1 IDEG $ideg} $aconds]
            append regcond { then 1 else 0 end }
            #puts regcond=$regcond

            namespace path [list {*}[namespace path] ::yacop::sectionizer]

            # try to make us interruptible
            set ::steenrod::_progvarname ::yacop::progVar
            yacop::interruptible {
                
                my db progress $::yacop::dbprogsteps yacop::ProgressHandler

                Section "Computing smash product" {
                    set lastflush [clock seconds]
                    $res db eval {
                        select rowid as resgen, sdeg as rsdeg, ideg as rideg, edeg as redeg
                        from generators order by sdeg,ideg,edeg
                    } {
                        my db eval $sql {

                            yacop::ProgressHandler

                            if {[my db onecolumn {
                                select count(*) from smash_generators 
                                where resgen=$resgen and modgen=$modgen and havediff limit 1
                            }]} continue
                            set ebasid [my db onecolumn {
                                select count(*) from smash_generators
                                where sdeg=$rsdeg+$msdeg 
                                and ideg=$rideg+$mideg 
                                and edeg=$redeg+$medeg
                            }]
                            my db eval {
                                insert or ignore into smash_generators
                                (resgen,modgen,sdeg,ideg,edeg,ebasid,havediff)
                                values
                                ($resgen,$modgen,$rsdeg+$msdeg,$rideg+$mideg,$redeg+$medeg,$ebasid,0)
                            }
                            my db eval {
                                select rowid as gid, sdeg,ideg,edeg from smash_generators 
                                where resgen=$resgen and modgen=$modgen
                            } break
                            my db eval {
                                insert or ignore into smash_boxes(sdeg,ideg,edeg)
                                values($sdeg,$ideg,$edeg)
                            }
                            # we can only compute differentials on this generator if 
                            # the receiving box (s-1,i,e) still lies within our region 
                            if {[my db onecolumn $regcond]} {
                                my MakeSmashFragments $gid
                            }
                            if {[clock seconds]>$lastflush+10} {
                                Section "Setting checkpoint at $rsdeg/$rideg/$redeg" {
                                    my db eval "commit;begin"
                                    set lastflush [clock seconds]
                                }
                            }
                        }
                        
                    }
                }
                Section "Computing homology" {

                    # TODO: fix where condition to make sure the resolution
                    # is complete in all required degrees

                    set sql [string map {
                        SDEG sdeg IDEG ideg
                        where {where homology is null and }
                    } [subst {
                        select rowid,sdeg,ideg,edeg from smash_boxes 
                        $wherecond order by sdeg,ideg,edeg 
                    }]]                    
                    #puts sql=$sql

                    my db eval $sql {
                        # we can only compute the homology if 
                        # the box (s-1,i,e) still lies within our region 
                        if {[my db onecolumn $regcond]} {                       
                            my MakeHomology $rowid
                            if {[clock seconds]>$lastflush+10} {
                                Section "Setting checkpoint at $sdeg/$ideg/$edeg" {
                                    my db eval "commit;begin"
                                    set lastflush [clock seconds]
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    method MakeSmashFragments {genid} {

        # we're only going to compute the portion of the differential
        # in the smash product which is chart- and homology-relevant.
        # these are the pts in the algebra and the unit
        
        lappend oplist {1 0 {} 0}
        lappend ptslist {}
        set p $rescfg(prime)
        set powers [steenrod::prime $p powers]
        foreach pts [steenrod::algebra ptslist $rescfg(algebra) [expr {[llength $powers]-1}]] {
            lappend ptslist $pts
            switch -- [lindex $pts 0] {
                P {
                    set x [string repeat "0 " [lindex $pts 1]]
                    lappend x [lindex $powers [lindex $pts 2]]
                    lappend oplist [list 1 0 [lrange $x 1 end] 0]
                }
                Q {
                    lappend oplist [list 1 [expr {1<<[lindex $pts 1]}] {} 0]
                }
            }
        }
        #puts oplist=$oplist

        my db eval {
            update smash_generators set havediff=1 where rowid = $genid;
            select resgen,modgen from smash_generators where rowid = $genid
        } break
        set mideg [my db onecolumn {select ideg from module_generators where rowid=$modgen}]
            
        # account for the internal differential on M, using d(m\land g) = dm\land g + ...
        foreach {cf tarmod} [$module diff $modgen] {
            my db eval {
                insert into 
                smash_fragments(srcgen,targen,opideg,format,data)
                values($genid,
                       (select rowid from smash_generators
                        where resgen=$targen and modgen=$tarmod),
                       0,'tcl','{{1 0 0 0}}');
            }
        }

        set modopdegs [my db eval {
            select distinct ideg-$mideg from module_generators where ideg >= $mideg 
        }]
        # go through all resolution fragments that originate on resgen,
        # then for all op from oplist create a corresponding smash 
        # fragment "op * (m*(a/op) smash $targen)" 
        $res db eval {
            select targen,data,opideg from fragview where srcgen = $resgen
        } {
            foreach pts $ptslist op $oplist {
                set ptsdeg [steenrod::mono degree $p $op]
                if {!( ($opideg-$ptsdeg) in $modopdegs )} continue
                set eshift 0
                if {[lindex $pts 0] eq "Q"} {
                    set eshift [lindex $op 1]
                }
                set sop $op
                lset sop 1 0
                catch {
                    lset sop 2 end -[lindex $sop 2 end]
                    # add a zero to mark this monomial as positive,
                    # despite the negative last entry
                    lset sop 2 "[lindex $sop 2] 0"
                }
                set shifted {}
                steenrod::poly foreach [steenrod::poly shift $data $sop] mono {
                    if {[steenrod::mono isabove $mono {1 0 0 0}]} {
                        # steenrod::poly shift doesn't handle the exterior
                        # part correctly if it is negative (we'll need a 
                        # special negshift for that)
                        foreach {cf ext red gen} $mono break
                        if {($eshift & $ext) == $eshift} {
                            if {$eshift} {
                                if {1 & bitcount(($eshift-1) & $ext)} {
                                    set cf [expr {-$cf}]
                                }
                                set ext [expr {$eshift ^ $ext}]
                            }
                            lappend shifted [list $cf $ext $red 0]
                        }
                    }
                }
                if {[llength $shifted]} {
                    set op2 $op
                    set action [$module actR $modgen $shifted]
                    #puts " $data/$op =  $shifted"
                    #puts " right action on $modgen: $action"

                    # we here deal with "pts * (m * (a/pts)) \tensor g"
                    # as a summand of d(m\tensor g)

                    if 0 {

                        we're here dealing with the summands

                        (-1)^{\vert m\vert} m \land (dg)

                        of the differential d(m\land g). we use

                        m\land ag = \sum (-1)^{\vert a'\vert\cdot \vert m\vert}
                        \cdot a'\left( \left(ma''\right)\land g\right)

                        here a' = pts, a'' = shifted

                    }

                    set sign [expr {($mideg*(1+$ptsdeg) & 1) ? -1 : 1}]

                    foreach {cf tarmod} $action {
                        lset op2 0 [expr {$sign*$cf}]
                        set op2l [list $op2]
                        my db eval {
                            insert into 
                            smash_fragments(srcgen,targen,opideg,format,data)
                            values($genid,
                                   (select rowid from smash_generators
                                    where resgen=$targen and modgen=$tarmod),
                                   $ptsdeg,'tcl',$op2l);
                        }
                    }
                }
            }
        }
    }

    method MakeHomology {boxid} {
        set p $rescfg(prime)
        my db eval {
            select sdeg,edeg,ideg from smash_boxes where rowid=$boxid;
            select cycles isnull as needc, boundaries isnull as needb 
            from smash_boxes where rowid=$boxid 
        } break
        if {$needc} {my MakeMatrix $sdeg $ideg $edeg}
        if {$needb} {my MakeMatrix [expr {$sdeg+1}] $ideg $edeg}
        my db eval {
            select cycles, boundaries from smash_boxes where rowid=$boxid 
        } break
        steenrod::matrix quot $p cycles $boundaries
        my db eval {
            update smash_boxes set homology=$cycles where rowid=$boxid
        }

        if {0 == [llength $cycles]} return

        set idx -1
        foreach row $cycles {
            incr idx
            my db eval {
                insert into homology_gendata(boxid,ebasid)
                values($boxid,$idx);
            }
            set hlgyrowid($idx) [my db last_insert_rowid]
        }

        my db eval {
            select rowid, ebasid,count(*) as srcdim from smash_generators
            where sdeg=$sdeg and ideg=$ideg and edeg=$edeg 
        } {
            set bid2gid($ebasid) $rowid
        }

        array set frags {}
        my db eval {
            select s.ebasid as srcbid, t.ebasid as tarbid, f.opideg as opideg, f.data as data
            from smash_generators s 
            join smash_fragments f on f.srcgen=s.rowid
            join smash_generators t on f.targen=t.rowid
            where s.sdeg=$sdeg and s.ideg=$ideg and s.edeg=$edeg
        } {
            set op [lindex $data 0]
            set cf [lindex $op 0]
            lset op 0 1
            lset op end 0
            lappend frags($op) $srcbid $tarbid $cf
            set fragdegs($op) $opideg
        }
        foreach {op deg} [array get fragdegs] {

            # only compute the well-defined multiplications, i.e. v_0 and h_k

            set isunit [expr {[lrange $op 1 2] eq "0 {}"}]
            set unwell [expr {[lindex $op 1] != 1 && [lindex $op 2 0] in "0 {}"}]

            if {$unwell} continue

            set eoff [expr {bitcount([lindex $op 1])}]
            set tardim [my db onecolumn {
                select count(*) from smash_generators 
                where sdeg=$sdeg-1 and ideg=$ideg-$deg and edeg=$edeg-$eoff
            }]
            set mat [steenrod::matrix create $srcdim $tardim]
            foreach {r c v} $frags($op) {
                set val [lindex $mat $r $c]
                set val [expr {($val+$v)%$p}]
                lset mat $r $c $val
            }
            set mat2 [steenrod::matrix multiply $p $cycles $mat]

            # express rows of mat2 in terms of homology generators

            my db eval {
                select homology, boundaries from smash_boxes 
                where sdeg=$sdeg-1 and ideg=$ideg-$deg and edeg=$edeg-$eoff 
            } break

            set hdim [llength $homology]
            if {$hdim == 0} continue

            set hcpy $homology
            lappend homology {*}$boundaries

            if {[llength $homology]} {            
                set lft [steenrod::matrix lift $p $homology mat2]
            } else {
                set lft [steenrod::matrix create {*}[steenrod::matrix dimensions $mat2]]
            }

            set idx -1
            foreach row $cycles lftrow $lft matrow $mat2 {
                incr idx
                if {[steenrod::matrix iszero [list $matrow]]} {
                    set cidx -1
                    foreach c $lftrow {
                        if {[incr cidx] >= $hdim} break
                        if {0 == $c} continue
                        set srcgen $hlgyrowid($idx)
                        set targen [my db onecolumn {
                            select rowid from homology_gendata
                            where boxid in (select rowid from smash_boxes 
                                            where sdeg=$sdeg-1 and ideg=$ideg-$deg and edeg=$edeg-$eoff)
                            and ebasid=$cidx
                        }]
                        if {$targen eq ""} {puts "$op:sdeg=$sdeg-1 and ideg=$ideg-$deg and edeg=$edeg-$eoff and ebasid=$cidx";barf;continue}
                        lset op 0 $c
                        my db eval {
                            insert into homology_fragments(srcgen,targen,format,data,opideg)
                            values($srcgen,$targen,'tcl',list($op),$deg);
                        }
                    }
                } else {           
                    puts \n\n\n\n
                    puts "$sdeg,$ideg,$edeg -> $op ($deg)"
                    puts $op:mat=$mat,cycles=$cycles,mat2=$mat2,homology=$hcpy,boundaries=$boundaries
                    puts srcdim=$srcdim,tardim=$tardim,eoff=$eoff,op=$op
                    puts lftrow=$lftrow,mat2row=$matrow
                    puts BARF!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                    #puts "$mat2cpy <-> $lft mit hcpy=$hcpy bnd=$boundaries"
                }
            }
        }


    }

    method MakeMatrix {sdeg ideg edeg} {
        # create matrix (s,i,e) -> (s-1,i,e)
        set p $rescfg(prime)
        set srcboxid [my db eval {
            select rowid from smash_boxes where sdeg=$sdeg and ideg=$ideg and edeg=$edeg
        }]
        set tarboxid [my db eval {
            select rowid from smash_boxes where sdeg=$sdeg-1 and ideg=$ideg and edeg=$edeg
        }]
        set srcdim [my db onecolumn {
            select count(*) from smash_generators where sdeg=$sdeg and ideg=$ideg and edeg=$edeg
        }]
        set tardim [my db onecolumn {
            select count(*) from smash_generators where sdeg=$sdeg-1 and ideg=$ideg and edeg=$edeg
        }]
        set matrix [steenrod::matrix create $srcdim $tardim]
        my db eval {
            select rowid as srcid, ebasid as srcebasid from smash_generators
            where sdeg=$sdeg and ideg=$ideg and edeg=$edeg order by ebasid asc
        } {
            my db eval {
                select targen,data,sg.ebasid as tarebasid from smash_fragments
                join smash_generators as sg on sg.rowid = targen
                where srcgen=$srcid and opideg=0
            } {
                set cf [lindex $data 0 0]
                set ncf [expr {($cf+[lindex $matrix $srcebasid $tarebasid])%$p}]
                lset matrix $srcebasid $tarebasid $ncf
            }
        }
        my db eval {
            update smash_boxes set matrix=$matrix where rowid=$srcboxid
        }
        steenrod::matrix ortho $p matrix ker 
        my db eval {
            update smash_boxes set cycles=$ker where rowid=$srcboxid
        }
        my db eval {
            update smash_boxes set boundaries=$matrix where rowid=$tarboxid
        }
    }
 
    method sage_homology_class {cycle} {
        set p $rescfg(prime)

        foreach {cf mname resgen} $cycle {
            my db eval {
                select sg.ebasid as ebasid, sb.rowid as boxid from 
                smash_generators as sg 
                join module_generators as mg on mg.rowid = sg.modgen
                join smash_boxes as sb
                on sb.sdeg = sg.sdeg and sb.ideg = sg.ideg and sb.edeg = sg.edeg
                where sg.resgen=$resgen and mg.sagename=$mname
            } {
                lappend vals($boxid) $cf $ebasid
            }
        }

        set res {}
        foreach box [array names vals] {
            my db eval {
                select homology, boundaries from smash_boxes where rowid=$box
            } break

            set hdim [llength $homology]
            lappend homology {*}$boundaries       
            
            foreach {r c} [steenrod::matrix dimension $homology] break
            set row [steenrod::matrix create 1 $c]

            #puts row=$row,vals=$vals($box),hom+bdr=$homology,hdim=$hdim

            foreach {cf idx} $vals($box) {
                set x [lindex $row 0 $idx]
                if {$x eq ""} {set x 0}
                set x [expr {($x+$cf)%$p}]
                lset row 0 $idx $x
            }

            if {[llength $homology]} {            
                set lft [steenrod::matrix lift $p $homology row]
            } else {
                set lft [steenrod::matrix create {*}[steenrod::matrix dimensions $row]]
            }
            
            if {![steenrod::matrix iszero $row]} {
                error "vector didn't reduce to zero"
            }

            set cidx -1
            foreach c [lindex $lft 0] {
                if {[incr cidx] >= $hdim} break
                if {0 == $c} continue
                set targen [my db onecolumn {
                    select rowid from homology_gendata
                    where boxid=$box and ebasid=$cidx
                }]
                if {$targen eq ""} {puts "$op:sdeg=$sdeg-1 and ideg=$ideg-$deg and edeg=$edeg-$eoff and ebasid=$cidx";barf;continue}
                lappend res "($c,$targen)"
            }
        }
        return "\[[join $res ,]\]"
    }
}