#
# Yacop startup file
#
# Copyright (C) 2009 Christian Nassau <nassau@nullhomotopie.de>
#
#  $Id: yacop.tcl,v 1.13 2009/05/31 10:59:53 cn Exp $
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#

# external requirements first

package require Tcl 8.5.7
package require TclOO 0.6
package require sqlite3 3.6.13
package require Steenrod 1.0

# source yacop files

foreach file {
    logging.tcl
    common.tcl
    prefs.tcl
    timer.tcl
    gfr.tcl
    smasher.tcl
    chart.tcl
    pschart.tcl
    chmap.tcl
} {
    source [file join [file dirname [info script]] $file]
}

# create a wrapper around sqlite which installs the default
# busy and progress handler:

proc yacop::sqlite {dbname {filename :memory:}} {   
    if {[string first : $filename] >= 0 && $filename ne ":memory:"} {
        error "colon in filename but not \":memory:\""
    }
    if {$filename ne ":memory:" && [info exists ::env(SAGE_DATA)]} {
        set filename [file join $::env(SAGE_DATA) yacop $filename]
    }
    if {$filename ne ":memory:"} {
        file mkdir [file dirname $filename]
    }
    uplevel 1 ::sqlite3 $dbname $filename   
    uplevel 1 $dbname busy ::yacop::sectionizer::BusyCB
    uplevel 1 $dbname progress $::yacop::dbprogsteps ::yacop::ProgressHandler
    foreach tclfunc {
        list lindex 
    } {
        uplevel 1 $dbname function $tclfunc ::$tclfunc
    }
    foreach cmd [info commands {::yacop::sqlitefunctions::[a-z]*}] {
        uplevel 1 $dbname function [namespace tail $cmd] $cmd
    }
    if {[package present yacop::sage]} {
        uplevel 1 [list ::yacop::sage::sqlitefuncs $dbname]
    }
}

package provide yacop 1.0
